@@ -74,9 +74,22 @@ class JobProcessor:
          self.terminating = True
 
 
+   def registerJobActionProcess(self,
+                                action):
+      actionProcessPID = os.getpid()
+      now              = time.time()
+
+      success,timeBetween = self.remoteJobMonitor.registerJobActionProcess(action,actionProcessPID,now)
+
+      return(success,timeBetween)
+
+
    def releaseRegisteredJobs(self):
       nReleasedJobs = self.remoteJobMonitor.releaseRegisteredJobs()
 
+      now = time.time()
+      self.remoteJobMonitor.updateJobActionProcessActivityTime('Processor',now)
+
       return(nReleasedJobs)
 
 
